/*============================================================================//
    PreferredSource.h: header file for PreferredSource UDT

    Purpose is to select a default solver --> check to ensure the binary 
     actually exists.  Note that this kind of defeats the purpose of using a
     generic Factory in SolverTraits (solvers register themselves), but we 
     cannot be sure that an actual executable exists.  Also, a compiled-in
     solver wrapper does not necessarily need to be a "preferred source".

    Implementation: Shane J. Neph, June 2004, University of Washington
//============================================================================*/

//Modified by Nathan Robinson, May 2008, Griffith University
// Copyright of the modifications (c) 2008 National ICT Australia Limited (NICTA)
//
// Author: Nathan Robinson (nathan.robinson@nicta.com.au)
//
// Redistribution and use in source and binary forms, with or without
// modification and only for non-commercial research and educational
// purposes are permitted provided that the conditions in the
// accompanying file "licence.txt" are met.


// Macro Guard
#ifndef PREFERRED_SOURCE_CSE473_H
#define PREFERRED_SOURCE_CSE473_H

// Files included
#include "ByLine.h"

#include "SolverInterface.h"
#include "SolverTraits.h"

struct PreferredSource {
    typedef SolverTraits::FactoryType FT;

    //NATHAN - removed most contents - now only checks for binaries
    static bool Exists(const std::string& dir, const std::string& solver) {
        return(exists(dir, solver)); // look for solver name directly
    }

    //NATHAN - removed contents
    static std::string SelectPreferred(const std::string& dir) {
        return("");
    }

private:
    static bool exists(const std::string& dir, const std::string& file) {
        std::string wholePath = file;
        if ( dir != "./" )
            wholePath = dir + wholePath;
        std::ifstream test(wholePath.c_str());
        if ( !test )
            return(false);
        std::vector<std::string> toCheck;
        std::copy(std::istream_iterator<ByLine>(test),
                  std::istream_iterator<ByLine>(),
                  std::back_inserter(toCheck));
        return(!toCheck.empty());
    }
};

#endif // PREFERRED_SOURCE_CSE473_H

