

/*********************************************************************
 * (C) Copyright 1999 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 *********************************************************************/

/*********************************************************************
 * File: main.c
 * Description: The main routine for the Branch-and-Bound Planner.
 *
 * Author: Joerg Hoffmann 2002
 *  Modified 2004 and 2006 
 *********************************************************************/ 

//Modified by Nathan Robinson, May 2008, Griffith University
// Copyright of the modifications (c) 2008 National ICT Australia Limited (NICTA)
//
// Author: Nathan Robinson (nathan.robinson@nicta.com.au)
//
// Redistribution and use in source and binary forms, with or without
// modification and only for non-commercial research and educational
// purposes are permitted provided that the conditions in the
// accompanying file "licence.txt" are met.

#include "bb.hh"
#include "memory.hh"
#include "output.hh"
#include "instantiateI.hh"
#include "instantiateII.hh"
#include "graph.hh"
#include "cnfout.hh"



#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <string.h>

extern int yyparse();



extern void initialiseLexer(istream*, ostream* );
extern void deleteLexer();

/*
 *  ----------------------------- GLOBAL VARIABLES ----------------------------
 */
FILE *yyin;
char* yytext;
/*******************
 * GENERAL HELPERS *
 *******************/

/* used to time the different stages of the planner
 */
float gtempl_time = 0, greach_time = 0, grelev_time = 0, gconn_time = 0;
float gbuild_time = 0, gsearch_time = 0, gcnf_time = 0;


/* the command line inputs
 */
struct _command_line gcmd_line;

/*

// ********
//PARSING
// ********

//NATHAN
//The roles of the following variables have been replaced by the
//new parser

//used for pddl parsing, flex only allows global variables
int gbracket_count;
char *gproblem_name;

//The current input line number
int lineno = 1;

//The current input filename
char *gact_filename;

//The pddl domain name
char *gdomain_name = NULL;

//loaded, uninstantiated operators
PlOperator *gloaded_ops = NULL;

//stores initials as fact_list 
PlNode *gorig_initial_facts = NULL;

//not yet preprocessed goal facts
PlNode *gorig_goal_facts = NULL;

//axioms as in UCPOP before being changed to ops
PlOperator *gloaded_axioms = NULL;

//to store all typed objects 
FactList *gorig_constant_list = NULL;

//type hierarchy (PDDL) 
type_tree_list gglobal_type_tree_list;

//helper for types parsing
FactList *gtypes = NULL;

//the predicates and their types as defined in the domain file
FactList *gpredicates_and_types = NULL;

*/

/*****************
 * INSTANTIATING *
 *****************/

/* global arrays of constant names,
 *               type names (with their constants),
 *               predicate names,
 *               predicate aritys,
 *               defined types of predicate args
 */
    
Token gconstants[MAX_CONSTANTS];           //the constants
int gnum_constants = 0;                    //the number of constants
Token gtype_names[MAX_TYPES];              //names of types
int gtype_consts[MAX_TYPES][MAX_TYPE];     //constants of type
Bool gis_member[MAX_CONSTANTS][MAX_TYPES]; //constant is member of type
int gtype_size[MAX_TYPES];                 //the size of the second dim of gtype_consts
int gnum_types = 0;                        //the number of types

Token gpredicates[MAX_PREDICATES];         //the names of the predicates
int garity[MAX_PREDICATES];                //the arity of the predicates
int gpredicates_args_type[MAX_PREDICATES][MAX_ARITY]; //the type of the pred args
int gnum_predicates = 0;                   //the number of predicates
    



/* the domain in integer (Fact) representation
 */
Operator_pointer goperators[MAX_OPERATORS];
int gnum_operators;
Fact gfull_initial[MAX_INITIAL];
int gnum_full_initial = 0;
Fact ggoal[MAX_STATE];
int gnum_goal = 0;


/* stores inertia - information: is any occurence of the predicate
 * added / deleted in the uninstantiated ops ?
 */
Bool gis_added[MAX_PREDICATES];
Bool gis_deleted[MAX_PREDICATES];

/* splitted initial state:
 * initial non static facts,
 * initial inertia facts of arity > 1
 */
Fact ginitial[MAX_STATE];
int gnum_initial = 0;
Fact ginertia[MAX_INITIAL];
int gnum_inertia = 0;

/* the type numbers corresponding to any unary inertia
 */
int gtype_to_predicate[MAX_PREDICATES];
int gpredicate_to_type[MAX_TYPES];

/* (ordered) numbers of types that new type is intersection of
 */
TypeArray gintersected_types[MAX_TYPES];
int gnum_intersected_types[MAX_TYPES];

/* intermediate step: store minimal action info, use it
 * to determine relevant facts, number of actions and final form
 * before creating actual actions
 */
ActionTemplate *gtemplates = NULL;          //These are not currently filled... unness
int gnum_templates = 0;

/* store the final "relevant facts"
 */
Fact grelevant_facts[MAX_RELEVANT_FACTS];
int gnum_relevant_facts = 0;
int gnum_pp_facts;


/* the (fully instantiated) domain in integer 
 * (number of relevant Fact) representation
 */
Action *gactions = NULL;
int gnum_actions;
State ginitial_state;
State ggoal_state;


/**********************
 * CONNECTIVITY GRAPH *
 **********************/

/* one fact array, ..
 */
FtConn *gft_conn;
int gnum_ft_conn;

/* one ops (actions) array
 */
OpConn *gop_conn;
int gnum_op_conn;

int gcword_size; 
int gnum_ft_bit;
int gnum_op_bit_at[MAX_GRAPH];
int gnum_ops = 0, gmax_block = 0;

/* the goals in bit-format...
 */
BitVector *gbit_goal_state;


/* some technical graph... things
 */


/* the ops not yet in the graph
 */
IntList *gout_ops = NULL;

/* global op-state resp. ft-state lists
 */ 
IntList *gin_ops = NULL;
IntList *gin_prev_ops = NULL;
IntList *gin_fts = NULL;
IntList *gin_prev_fts = NULL;

/* we want to have fast access, during constraint propagation,
 * to the fts that can (maximally) be present at the layer;
 * can easily store this by means of pointers into the gin_fts list
 * which we got anyway.
 */
IntList *gin_fts_at[MAX_GRAPH];
/* the present ops are good to have for enforcing a positive constraint:
 * we must then negatively enforce the exclusive ops. don't wanna store
 * all the exclusives in a list but at least we can skip those ops that
 * aren't even there. means as above.
 */
IntList *gin_ops_at[MAX_GRAPH];

/* current mutexes: exclusives speedup
 */
IntPair *gin_op_mutex_pairs = NULL;
IntPair *gin_ft_mutex_pairs = NULL;

/* information about current state of graph, needed for level off test
 */
int gin_ft_count = 0, gin_ft_exclusion_count = 0;
int gin_op_count = 0, gin_op_exclusion_count = 0;

/* is TRUE iff graph has levelled off.
 */
Bool gsame_as_prev_flag = FALSE;

/* stores the time step at which graph has levelled off.
 */
int gfirst_full_time = 0;

/* some search things
 */
 
/* flag saying that depth first has encountered a solution!
 */
Bool gfound_plan;

/* number of action constraints enforced successfully
 */
int gconstr_succ = 0;
/* same, unsuccessfully
 */
int gconstr_fail = 0;

/* search info: the currently positively enforced ops at each time
 */
int **genforced_ops_at;
int *gnum_enforced_ops_at;

/* during PG re-computation: forcing an action A in might
 * force actions below in, too - namely, if they're the
 * only achievers of a prec of A. remember these actions and
 * force them in once propagation for A has terminated.
 */
Decvar gto_be_enforced[ARRAY_SIZE];
int gnum_to_be_enforced;


/* some statistics on transitively forced in - ops
 * names refer to paper work on unit propag == rplan propag
 */
int gnum_bins = 0;
int gnum_d2ins = 0;

struct tms gstart, gend;

/* the dec vars sufficing for BBG search at the respective layers!
 * 
 * in case of solution: dec vars below that sol.
 * in case of unsat:  all dec vars expanded during search (?)
 */
Decvar gbbg_decvars[MAX_GRAPH][MAX_DECVARS];
int gnum_bbg_decvars[MAX_GRAPH];
int gnum_bbg_decs;

/* this is what bd verification works on.
 */
Decvar gbd_candidate[MAX_DECVARS];
int gnum_bd_candidate;



/* technical, for backdoor verification.
 */
Bool gsolution_found;
Bool girresolved_found;
int gbd_candidate_curr;
Bool gcontrol_strongbd;
 


/* rplan now also needed in backdoors.
 */
int **grelaxed_ops_at;
int *gnum_relaxed_ops_at;



/* the decvar Database to choose from, for 
 * enumeration
 */
Decvar gD[MAX_D_DECVARS];
int gnum_D;


/*
 *  ----------------------------- HEADERS FOR PARSING ----------------------------
 * ( fns defined in the scan-* files )
 */


void get_fct_file_name( char *filename );
void load_ops_file( char *filename );
void load_fct_file( char *filename );



struct tms start, end;

/*
 *  ----------------------------- MAIN ROUTINE ----------------------------
 */

/* start - globals and prototypes added by sjn */
const int SAT = 0;
const int UNSAT = 2;
const int UNRECOVERABLEERROR = 1;
int get_solution(FILE* fp, FILE* solnFile);
int create_solution();
extern int *lop_to_code;
extern int **lcode;
extern int *ltime_to_code;
/* end */

int lh2level, lreallevel;

int fct_pddlwraw = 1;


//This function was added by NATHAN
bool local_parse_function(const string &domainFileName, const string &problemFileName)
{
    ifstream domainFile;
    domainFile.open(domainFileName.c_str(), ifstream::in);
    if(!domainFile.is_open()){
	    cerr << "Couldn't open domain definition.\n";
	    return false;
    }
    initialiseLexer(&domainFile, &cout);
    if(yyparse()){
	    cerr << "Parsing of domain file :: "<<domainFileName<<" failed!\n";
	    return false;
    }
    deleteLexer();
    domainFile.close();
    ifstream problemFile;
    problemFile.open(problemFileName.c_str(), ifstream::in);
    if(!problemFile.is_open()){
	    cerr << "Couldn't open problem definition.\n";
	    return false;
    }
    initialiseLexer(&problemFile, &cout);
    if(yyparse()){
	    cerr << "Parsing of problem file :: "<<problemFileName<<" failed!\n";
	    return false;
    }
    deleteLexer();
    problemFile.close();

    /*Types*/
    problem.domain.configureTypes();
    problem.domain.unwindTypes();
    /*Constants*/
    problem.domain.configureConstants();
    problem.domain.unwindConstants();
    /*Constants -- objects*/
    problem.configureObjects();
    problem.unwindObjects();

    problem.domain.initialiseMirrors();
    problem.initialiseMirrors();

    /*Decide what predicate symbols are fluent, and their type of fluency*/
    problem.domain.processActionsForSymbolTypes();
    
    return true;
} 


//MODIFIED by NATHAN
int main( int argc, char *argv[] )

{
  /* resulting name for ops file
   */
  char ops_file[MAX_LENGTH] = "";
  /* same for fct file 
   */
  char fct_file[MAX_LENGTH] = "";
  
  gcmd_line.cnflayer = -1;
  gcmd_line.cnfout = -1;
  gcmd_line.solverOut = FALSE;
  gcmd_line.prune = FALSE;
  gcmd_line.binary_clause_only = 0;
  strcpy(gcmd_line.cnfFileName, "CNF");
  gcmd_line.makeCNF = 0;
  gcmd_line.prevBestScore = -1;


  /* command line treatment
   */
  if ( argc == 1 || ( argc == 2 && *++argv[0] == '?' ) ) {
    bb_usage();
    exit( 1 );
  }
  if ( !process_command_line( argc, argv ) ) {
    printf("\n\nEXIT: bad arguments detected\n\n");
    bb_usage();    
    exit( 1 );
  }


  /* make file names
   */

  /* one input name missing
   */
  if ( !gcmd_line.ops_file_name || 
       !gcmd_line.fct_file_name ) {
    fprintf(stdout, "\nbb: two input files needed\n\n");
    bb_usage();      
    exit( 1 );
  }
  /* add path info, complete file names will be stored in
   * ops_file and fct_file 
   */
  sprintf(ops_file, "%s%s", gcmd_line.path, gcmd_line.ops_file_name);
  sprintf(fct_file, "%s%s", gcmd_line.path, gcmd_line.fct_file_name);

  /* parse the input files
   */

  /* start parse & instantiation timing
   */
  times( &start );
  times( &gstart );
  /* domain file (ops)
   */
   
  //The following parsing code was added by NATHAN
  local_parse_function(string(ops_file), string(fct_file));
  Planning::CostPlanner planner(problem, 0);
  // now instantiate operators; 
  // start by collecting all strings and thereby encoding 
  // the domain in integers.
  encode_domain_in_integers(planner);
   // inertia preprocessing:
   //   - collect inertia information
   //   - split initial state into
   //        - types for unary inertia
   //        - array containing all inertia relations
   //        - array containing non - static relations
   //   - encode unary inertia in op preconds into types
   //this has been modified... 
    do_inertia_preprocessing();
   // unify inertia preconds with initial state,
   // then multiply remaining parameters;
   // create one ActionTemplate for each possible parameter combination
   //
   // --- the actual instantiation part, creating all possible instances ---
  //build_action_templates();
  //instead of the above::
  //Copy the data structures from charles parser into satplans memory
  copy_action_templates_from_parser(planner);


  times(&end);
  TIME( gtempl_time );

  times(&start);

   // perform a simple reachabilty analysis
   // (added facts fixpoint)
   // to find out which facts can (at most) be made true in domain
   //
   perform_reachability_analysis();

  times(&end);
  
  
  TIME( greach_time );

  times(&start);

  /* as a last step, collect relevant facts (deleted ini's and
   * added non ini's) and remove others from domain
   */
  collect_relevant_facts();



  times(&end);
  TIME( grelev_time );

  times(&start);

  /* now build globally accessible connectivity graph
   */
  build_connectivity_graph();


  /* insert bit vectors for fast lookup
   *
   * (NOTE: bit vector sizes are static across graph here;
   *  (in difference to IPP) -- otherwise pre-computation 
   *  not possible as done in these parts)
   */
  insert_bit_vectors();
 

  times(&end);
  TIME( gconn_time );

  /* output CNF file */
  times(&start);
  do_cnf_output(gcmd_line.prevBestScore);
  
  //No longer needed (NATHAN)
  //free_PlOperator( gloaded_ops );
  
  times(&end);
  TIME ( gcnf_time );
  
  output_planner_info();

 /* if ( gcmd_line.makeCNF != 0 )*/ /* output solution file */
      /*exit(create_solution());*/
  exit( SAT );
}


/*
 *  ----------------------------- HELPING FUNCTIONS ----------------------------
 */


void output_planner_info( void )

{
  //Removed by NATHAN
  //printf( "\n\ntime spent: %7.2f seconds instantiating %d action templates", 
//	  gtempl_time, gnum_templates );
  //printf( "\n            %7.2f seconds reachability analysis, yielding %d facts and %d actions", 
	//  greach_time, gnum_pp_facts, gnum_actions );
  //printf( "\n            %7.2f seconds collecting %d relevant facts", 
//	  grelev_time, gnum_relevant_facts );
  //printf( "\n            %7.2f seconds building connectivity graph", gconn_time );
  //printf( "\n            %7.2f seconds building (std) graph", gbuild_time );
  //printf( "\n            %7.2f seconds CNF output time", gcnf_time );
  //printf( "BB time: %7.2f\n", 
//	  gtempl_time + greach_time + grelev_time + gconn_time + gbuild_time + gcnf_time);
}

void output_planner_info_file(FILE* fp, int numSteps) {
    fprintf(fp, "; ParsingTime %7.2f\n", gbuild_time);
    fprintf(fp, "; MakeSpan %d\n", numSteps);
}

void bb_usage( void )

{

  printf("\nusage of bb:\n");

  printf("\nOPTIONS   DESCRIPTIONS\n\n");
  printf("-p <str>    path for operator and fact file\n");
  printf("-o <str>    operator file name\n");
  printf("-f <str>    fact file name\n\n");

  printf("-l  <num>   goal layer for CNF\n\n");

  printf("-G <0 or 1> (0) create CNF output or (1) build final solution\n");
  printf("-b <str>    CNF output file name\n");
  printf("-t <0 or 1> (1) CNF output includes only unary/binary clauses - others ignored\n");
  printf("-S <str>    Input Solution File Name (only when -G 1 is used)\n");
  printf("-F <str>    Final Output Solution File Name (only when -G 1 is used)\n");
  printf("-V <str>    Variables File Name - list all variables (only when -G 1 is used)\n");

  printf("-C          CNF formula output (preset: %d); at layer <-l>\n",
	 gcmd_line.cnfout);
  printf("      0     none\n");
  printf("      1     action-based\n");
  printf("      2     gp-style action-based\n");
  printf("      3     gp-based\n");
  printf("      4     thin gp-based\n\n");


}

//Modified by NATHAN (added prevBestScore)
Bool process_command_line( int argc, char *argv[] )

{

  char option;


  gcmd_line.display_info = 1;
  gcmd_line.debug = 0;
  gcmd_line.min_time = 0;
  
  memset(gcmd_line.ops_file_name, 0, MAX_LENGTH);
  memset(gcmd_line.fct_file_name, 0, MAX_LENGTH);
  memset(gcmd_line.path, 0, MAX_LENGTH);

  while ( --argc && ++argv ) {
    if ( *argv[0] != '-' || strlen(*argv) != 2 ) {
	return FALSE;
    }
    option = *++argv[0];
    switch ( option ) {
    default:
      if ( --argc && ++argv ) {
	switch ( option ) {
	case 'p':
	  strncpy( gcmd_line.path, *argv, MAX_LENGTH );
	  break;
	case 'o':
	  strncpy( gcmd_line.ops_file_name, *argv, MAX_LENGTH );
	  break;
	case 'f':
	  strncpy( gcmd_line.fct_file_name, *argv, MAX_LENGTH );
	  break;
	case 'i':
	  sscanf( *argv, "%d", &gcmd_line.display_info );
	  break;
	case 'd':
	  sscanf( *argv, "%d", &gcmd_line.debug );
	  break;
	case 'm':
	  sscanf( *argv, "%d", &gcmd_line.min_time );
	  break;
	case 'l':
	  sscanf( *argv, "%d", &gcmd_line.cnflayer );
	  break;
	case 'z':
	  sscanf( *argv, "%d", &gcmd_line.prevBestScore );
	  break;
	case 'C':
	  sscanf( *argv, "%d", &gcmd_line.cnfout );
	  break;
	case 's':
	  gcmd_line.solverOut = TRUE;
	  ++argc;
	  --argv;
	  break;
        case 't':
          sscanf( *argv, "%d", &gcmd_line.binary_clause_only);
          break;
	case 'P':
	  gcmd_line.prune = TRUE;
	  ++argc;
	  --argv;
	  break;
        case 'b':
          strncpy( gcmd_line.cnfFileName, *argv, MAX_LENGTH );
          break;
        case 'G':
          sscanf( *argv, "%d", &gcmd_line.makeCNF );
          break;
        case 'S':
	  strncpy( gcmd_line.input_solution, *argv, MAX_LENGTH );
	  break;
        case 'F':
	  strncpy( gcmd_line.final_solution, *argv, MAX_LENGTH );
	  break;
        case 'V':
	  strncpy( gcmd_line.varFileName, *argv, MAX_LENGTH );
          break;
	default:
	  printf( "\nbb: unknown option: %c entered\n\n", option );
	  return FALSE;
	}
      } else {
	return FALSE;
      }
    }
  }

  if ( gcmd_line.cnflayer < 1 ) {
    return FALSE;
  }
       

  return TRUE;

}

