
/*********************************************************************
 * (C) Copyright 1999 Albert Ludwigs University Freiburg
 *     Institute of Computer Science
 *
 * All rights reserved. Use of this software is permitted for 
 * non-commercial research purposes, and it may be copied only 
 * for that use.  All copies must include this copyright message.
 * This software is made available AS IS, and neither the authors
 * nor the  Albert Ludwigs University Freiburg make any warranty
 * about the software or its performance. 
 *********************************************************************/
 
/*********************************************************************
 * File: backdoors.c (2004) and cnfout.c (2006)
 * Description: functions for cnf output
 *
 * Author: Joerg Hoffmann 2003
 *  Modified: Shane Neph 2004 and 2006
 *********************************************************************/ 
 
//Modified by Nathan Robinson, May 2008, Griffith University
// Copyright of the modifications (c) 2008 National ICT Australia Limited (NICTA)
//
// Author: Nathan Robinson (nathan.robinson@nicta.com.au)
//
// Redistribution and use in source and binary forms, with or without
// modification and only for non-commercial research and educational
// purposes are permitted provided that the conditions in the
// accompanying file "licence.txt" are met.

#include "cnfout.hh"

int cctr;

int **lcode;

int **lclause;
int *lclause_size;
int lnum_clauses;

//Added by NATHAN
#define TYPE_PROP 0
#define TYPE_ACTION 1
#define TYPE_NOOP 2


//added by NATHAN
vector<int>    code_type;
vector<int>    code_opt_or_fact;
vector<int>    code_to_time;
vector<int>    code_to_score;
vector<string> code_to_name;


//Modified by NATHAN
void do_cnf_output(int prevBestScore)
{
	print_gp_based_encoding( gcmd_line.cnflayer, prevBestScore );
}


//NATHAN
//The variable naming portion of this code has been changed.
//The outputting of the cnf into files has also been changed.
void print_gp_based_encoding( int layer, int prevBestScore )

{
    int i, j, t, ft, op, code;
    IntList *tmp, *i1, *i2;

    //have to separate the ft coding out from the op coding.
    int **fcode;

    lclause = ( int ** ) calloc( MAX_CLAUSES, sizeof( int * ) );
    lclause_size = ( int * ) calloc( MAX_CLAUSES, sizeof( int ) );
    lcode = ( int ** ) calloc( layer, sizeof( int * ) );
    
    
    for ( i = 0; i < layer; i++ ) {
    	lcode[i] = ( int * ) calloc( gnum_op_conn, sizeof( int ) );
    	for ( j = 0; j < gnum_op_conn; j++ ) {
    	    lcode[i][j] = -1;
    	}
    }
    fcode = ( int ** ) calloc( layer + 1, sizeof( int * ) );
    for ( i = 0; i <= layer; i++ ) {
    	fcode[i] = ( int * ) calloc( gnum_ft_conn, sizeof( int ) );
    	for ( j = 0; j < gnum_ft_conn; j++ ) {
    	    fcode[i][j] = -1;
    	}
    }

    //build the graph up to <layer>
    for ( i = 0; i < gnum_op_conn; i++ ) {
    	tmp = new_IntList( i );
    	if ( gout_ops ) {
    	    gout_ops->prev = tmp;
    	}
    	tmp->next = gout_ops;
    	gout_ops = tmp;
    }
    for ( i = 0; i < ginitial_state.num_F; i++ ) {
    	ft = ginitial_state.F[i];
    	gin_ft_count++;
    	gft_conn[ft].first_appearance = 0;
    	gft_conn[ft].info_at[0] = new_FtLevelInfo();
    	tmp = new_IntList( ft );
    	tmp->next = gin_fts;
    	gin_fts = tmp;
    }
    gin_fts_at[0] = gin_fts;
    //if ( gcmd_line.display_info ) {
    	//printf("\ntime: %3d, %5d facts and %7d exclusive pairs",
    	//       0, gin_ft_count, gin_ft_exclusion_count);
    	//fflush(stdout);
    //}
    for( t = 0; t < layer; t++ ) {
    	build_graph_evolution_step();
    	fflush(stdout);
    }

    //prec constraints, + code assignment.
    lnum_clauses = 0;
    code = 1; // variables numbered 1 .. n 
    
    code_type.push_back(TYPE_PROP);
    code_opt_or_fact.push_back(-1);
    code_to_time.push_back(-1);
    code_to_score.push_back(-1);
    code_to_name.push_back("");
    
    for ( t = 0; t <= layer; t++ ) {
        //printf("plan graph layer %d...\n", t);
        for ( i = 0; i < gnum_ft_conn; i++ ) {
            if ( gft_conn[i].info_at[t] == NULL) continue;
            
            // insert adder clause, if at t > 0. else, say it has to be true,
            // if it's an initial fact.
            if ( t == 0 ) {
                for ( j = 0; j < ginitial_state.num_F; j++ ) {
                    if ( ginitial_state.F[j] == i ) break;
                }
                if ( j < ginitial_state.num_F ) {
                    fcode[t][i] = code++;
                    
                    code_type.push_back(TYPE_PROP);
                    code_opt_or_fact.push_back(i);
                    code_to_time.push_back(t);
                    code_to_score.push_back(0);
                    code_to_name.push_back(print_FactToString(&(grelevant_facts[i])) + " " + toString(t));

                    if ( lnum_clauses == MAX_CLAUSES ) {
                        printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                        exit( 1 );
                    }
                    lclause[lnum_clauses] = ( int * ) calloc( 1, sizeof( int ) );
                    lclause[lnum_clauses][0] = fcode[t][i];
                    lclause_size[lnum_clauses] = 1;
                }
                continue;
            }
            
            if ( fcode[t][i] == -1 ) {
                fcode[t][i] = code++;
                
                code_type.push_back(TYPE_PROP);
                code_opt_or_fact.push_back(i);
                code_to_time.push_back(t);
                code_to_score.push_back(0);
                code_to_name.push_back(print_FactToString(&(grelevant_facts[i])) + " " + toString(t));
            }
      
            if ( lnum_clauses == MAX_CLAUSES ) {
                printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                exit( 1 );
            }
            
            lclause[lnum_clauses] = ( int * ) calloc( gft_conn[i].num_A+1, sizeof( int ) );
            lclause[lnum_clauses][0] = (-1) * fcode[t][i];
            lclause_size[lnum_clauses] = 1;
            for ( j = 0; j < gft_conn[i].num_A; j++ ) {
                op = gft_conn[i].A[j];
                if ( gop_conn[op].info_at[t-1] == NULL ) {
                    continue;
                }
                if ( lcode[t-1][op] == -1 ) {
                    lcode[t-1][op] = code++;
                    if ( 0 ) {
                        printf("\npC %d at %d", op, t-1);
                    }
                    if ( code == MAX_CNF_VARS + 1 ) {
                        printf("\n\ntoo many cnf vars. increase MAX_CNF_VARS.\n\n");
                        exit( 1 );
                    }
                   
                    if (gop_conn[op].noop_for == -1) code_type.push_back(TYPE_ACTION);
                    else code_type.push_back(TYPE_NOOP);
                    code_opt_or_fact.push_back(op);
                    code_to_time.push_back(t-1);
                    code_to_score.push_back(gop_conn[op].cost);
                    code_to_name.push_back(print_op_nameToString(op) + " " + toString(t-1));
                   
                }
                lclause[lnum_clauses][lclause_size[lnum_clauses]++] = lcode[t-1][op];
            }
            
            if ( lclause_size[lnum_clauses] == 1 ) {
                printf("\n\nno achiever in at t>0??\n\n");
                exit( UNSAT );
            }
            lnum_clauses++;
        } /* i-ft */

        if ( t == layer ) {
            continue;
        }

       
        for ( i = 0; i < gnum_op_conn; i++ ) {
            if ( gop_conn[i].info_at[t] == NULL) {
                continue;
            }
            // insert prec clauses.
            for ( j = 0; j < gop_conn[i].num_P; j++ ) {
                ft =  gop_conn[i].P[j];

                if ( lnum_clauses == MAX_CLAUSES ) {
                    printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
                //here the op i is used at t, for the 1st time. assign the code.
                if ( j == 0 && lcode[t][i] == -1) {
                    lcode[t][i] = code++;

                    if (gop_conn[i].noop_for == -1) code_type.push_back(TYPE_ACTION);
                    else code_type.push_back(TYPE_NOOP);
                    code_opt_or_fact.push_back(i);
                    code_to_time.push_back(t);
                    code_to_score.push_back(gop_conn[i].cost);
                    code_to_name.push_back(print_op_nameToString(i) + " " + toString(t));
                }

                lclause[lnum_clauses][0] = (-1) * lcode[t][i];
                if ( fcode[t][ft] == -1 ) {
                    printf("\n\nprec fact not coded??\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses][1] = fcode[t][ft];
                lclause_size[lnum_clauses] = 2;
                lnum_clauses++;
            } // j: pres of i-op 
            
            
            //check if something is both added and deleted
            set<int> addF, ignoreEff;
            for ( j = 0; j < gop_conn[i].num_A; j++ ) {
                addF.insert(gop_conn[i].A[j]);
            }
            for ( j = 0; j < gop_conn[i].num_D; j++ ) {
                ft = gop_conn[i].D[j];
                if (addF.find(ft) != addF.end()) {
                    ignoreEff.insert(ft);
                }
            }
            
            //Add effects
            for ( j = 0; j < gop_conn[i].num_A; j++ ) {
                ft =  gop_conn[i].A[j];
                if (ignoreEff.find(ft) != ignoreEff.end()) continue;

                if ( lnum_clauses == MAX_CLAUSES ) {
                    printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
                
                if ( j == 0 && lcode[t][i] == -1) {
                    lcode[t][i] = code++;
                    if (gop_conn[i].noop_for == -1) code_type.push_back(TYPE_ACTION);
                    else code_type.push_back(TYPE_NOOP);
                    code_opt_or_fact.push_back(i);
                    code_to_time.push_back(t);
                    code_to_score.push_back(gop_conn[i].cost);
                    code_to_name.push_back(print_op_nameToString(i) + " " + toString(t));
                }
                lclause[lnum_clauses][0] = (-1) * lcode[t][i];
                
                if (fcode[t+1][ft] == -1) {
                    fcode[t+1][ft] = code++;
                    code_type.push_back(TYPE_PROP);
                    code_opt_or_fact.push_back(ft);
                    code_to_time.push_back(t+1);
                    code_to_score.push_back(0);
                    code_to_name.push_back(print_FactToString(&(grelevant_facts[ft])) + " " + toString(t+1));
                    
                }
                lclause[lnum_clauses][1] = fcode[t+1][ft];
                lclause_size[lnum_clauses] = 2;
                lnum_clauses++;
            }
            
        } // i -- op 
        
        //for delete effects
        for ( i = 0; i < gnum_op_conn; i++ ) {
            if ( gop_conn[i].info_at[t] == NULL) {
                continue;
            }
            //check if something is both added and deleted
            set<int> addF, ignoreEff;
            for ( j = 0; j < gop_conn[i].num_A; j++ ) {
                addF.insert(gop_conn[i].A[j]);
            }
            for ( j = 0; j < gop_conn[i].num_D; j++ ) {
                ft = gop_conn[i].D[j];
                if (addF.find(ft) != addF.end()) {
                    ignoreEff.insert(ft);
                }
            }
            //Delete effects
            for ( j = 0; j < gop_conn[i].num_D; j++ ) {
                ft = gop_conn[i].D[j];
                if (ignoreEff.find(ft) != ignoreEff.end()) continue;
                if (fcode[t+1][ft] == -1) continue; 
                if ( lnum_clauses == MAX_CLAUSES ) {
                    printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
                
                if ( j == 0 && lcode[t][i] == -1) {
                    lcode[t][i] = code++;
                    if (gop_conn[i].noop_for == -1) code_type.push_back(TYPE_ACTION);
                    else code_type.push_back(TYPE_NOOP);
                    code_opt_or_fact.push_back(i);
                    code_to_time.push_back(t);
                    code_to_score.push_back(gop_conn[i].cost);
                    code_to_name.push_back(print_op_nameToString(i) + " " + toString(t));
                }
                lclause[lnum_clauses][0] = (-1) * lcode[t][i];
                
                lclause[lnum_clauses][1] = (-1) * fcode[t+1][ft];
                lclause_size[lnum_clauses] = 2;
                lnum_clauses++;
            }
            
        } // i -- op 
    } // t 


    // goal constraints
    //printf("\ngoal constraints...");
    for ( i = 0; i < ggoal_state.num_F; i++ ) {
        ft = ggoal_state.F[i];

        if ( lnum_clauses == MAX_CLAUSES ) {
            printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
            exit( 1 );
        }
        lclause[lnum_clauses] = ( int * ) calloc( 1, sizeof( int ) );
        if ( fcode[layer][ft] == -1 ) {
            //printf("\n\ngoal fact not coded - level too low??\n\n");
            exit(UNSAT);
        }
        lclause[lnum_clauses][0] = fcode[layer][ft];
        lclause_size[lnum_clauses] = 1;
        lnum_clauses++;
    }


    // exclusion constraints.
    for ( t = 0; t <= layer; t++ ) {
        //printf("\nexclusion constraints layer %d...", t);
        for ( i1 = gin_fts_at[t]; i1 && i1->next; i1 = i1->next ) {
            i = i1->i1;
            for ( i2 = i1->next; i2; i2 = i2->next ) {
                j = i2->i1;
                if ( !(gft_conn[i].info_at[t]->bit_exclusives[gft_conn[j].uid_block] &
                    gft_conn[j].uid_mask) ) {
                    continue;
                }
                if ( lnum_clauses == MAX_CLAUSES ) {
                    printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
                if ( fcode[t][i] == -1 ) {
                    fcode[t][i] = code++;
                    
                    code_type.push_back(TYPE_PROP);
                    code_opt_or_fact.push_back(i);
                    code_to_time.push_back(t);
                    code_to_score.push_back(0);
                    code_to_name.push_back(print_FactToString(&(grelevant_facts[i])) +
                        " " + toString(t));
                }
                lclause[lnum_clauses][0] = (-1) * fcode[t][i];
                if ( fcode[t][j] == -1 ) {
                    fcode[t][j] = code++;
                    
                    code_type.push_back(TYPE_PROP);
                    code_opt_or_fact.push_back(j);
                    code_to_time.push_back(t);
                    code_to_score.push_back(0);
                    code_to_name.push_back(print_FactToString(&(grelevant_facts[j])) +
                        " " + toString(t));
                }
                lclause[lnum_clauses][1] = (-1) * fcode[t][j];
                lclause_size[lnum_clauses] = 2;
                lnum_clauses++;
            } // i2 
        } // i1 

        if ( t == layer ) continue;

        for ( i1 = gin_ops_at[t]; i1 && i1->next; i1 = i1->next ) {
            i = i1->i1;
            for ( i2 = i1->next; i2; i2 = i2->next ) {
                j = i2->i1;
                if ( !(gop_conn[i].info_at[t]->bit_exclusives[gop_conn[j].uid_block] &
                    gop_conn[j].uid_mask) ) continue;
                
                if ( lnum_clauses == MAX_CLAUSES ) {
                    printf("\n\ntoo many clauses. increase MAX_CLAUSES.\n\n");
                    exit( 1 );
                }
                lclause[lnum_clauses] = ( int * ) calloc( 2, sizeof( int ) );
                if ( lcode[t][i] == -1 ) {
                    lcode[t][i] = code++;
                    
                    if (gop_conn[i].noop_for == -1) code_type.push_back(TYPE_ACTION);
                    else code_type.push_back(TYPE_NOOP);
                    code_opt_or_fact.push_back(i);
                    code_to_time.push_back(t);
                    code_to_score.push_back(gop_conn[i].cost);
                    code_to_name.push_back(print_op_nameToString(i) + " " + toString(t));
                }
                
                lclause[lnum_clauses][0] = (-1) * lcode[t][i];
                if ( lcode[t][j] == -1 ) {
                    lcode[t][j] = code++;
                    
                    if (gop_conn[j].noop_for == -1) code_type.push_back(TYPE_ACTION);
                    else code_type.push_back(TYPE_NOOP);
                    code_opt_or_fact.push_back(j);
                    code_to_time.push_back(t);
                    code_to_score.push_back(gop_conn[j].cost);
                    code_to_name.push_back(print_op_nameToString(j) + " " + toString(t));
                }
                lclause[lnum_clauses][1] = (-1) * lcode[t][j];
                lclause_size[lnum_clauses] = 2;
                lnum_clauses++;
            }
        }
    }
    
    //printf("DECISION LAYER %d, WRITING GP-BASED ENCODING WITH %d VARS, %d CLAUSES.\n", 
	//   layer, code-1, lnum_clauses);
    
    ofstream varFile;
    varFile.open(gcmd_line.varFileName);
    //cout << "varFileName: " << gcmd_line.varFileName << endl;
    if (!varFile.is_open()) {cerr << "Unable to open file: " << 
        string(gcmd_line.varFileName) << endl; exit(1);}
    varFile << (code-1) << endl;
    for (int vPos = 1; vPos < code; vPos++) {
        varFile << "v " << vPos << " ";
        if (code_type[vPos] == TYPE_ACTION) varFile << "a ";
        else if (code_type[vPos] == TYPE_NOOP) varFile << "n ";
        else if (code_type[vPos] == TYPE_PROP) varFile << "p ";
        else assert(0);
        varFile << code_to_time[vPos] << " " << code_to_score[vPos] <<
            " " << code_to_name[vPos] << " -1" << endl; 
        //Finally list all of effects of the actions...
        
    }
    varFile.close();
    
    ofstream cnfFile;
    cnfFile.open(gcmd_line.cnfFileName);
    //cout << "cnfFileName: " << gcmd_line.cnfFileName << endl;
    if (!cnfFile.is_open()) {cerr << "Unable to open file: " << 
        string(gcmd_line.cnfFileName) << endl; exit(1);}
    for (int vPos = 1; vPos < code; vPos++) {
        cnfFile << "c cv " << vPos << " " << code_to_score[vPos] << endl;
    }
    //NATHAN *11
    //if (prevBestScore > -1) {
    //    cnfFile << "b cost " << prevBestScore << endl;
    //}
    cnfFile << "p cnf " << code-1 << " " << lnum_clauses << endl;
    for ( i = 0; i < lnum_clauses; i++ ) {
	    for ( j = 0; j < lclause_size[i]; j++ ) {
	        cnfFile << lclause[i][j] << " ";
	    }
	    cnfFile << "0\n";
    }
    cnfFile.close();
}

